#define CRASH_PATH_PROFILE_LENGTH	1024

/*
 *	Atom program to record (and report) a block level path
 *	profile upon exit of the program
 *
 *	Purpose:
 *		Post-mortem diagnostic tool (which can provide better
 *		context than a simple stack traceback)
 *
 *	Written by Rick Gorton 22-Feb-2000
 *	gorton@world.std.com
 *
 *	Copyright (C) 2000, Rick Gorton
 *
 * This ATOM tool is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * This ATOM tool is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this ATOM tool; see the file COPYING.  If not, write to
 * the Free Software Foundation, 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 *
 * NOTE:
 *	If you find this tool useful, the author is interested in knowing
 *	how you used it, and what kind of modifications you found useful.
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include "datatypes.h"
#include <signal.h>

/*
 * Data declarations local to this file
 */
long	Total_Procs;
PROC_TABLE_TYPE		*dyn_proc_count;
long	total_proc_count;

long	Total_Blocks;
BLOCK_TABLE_TYPE	*dyn_block_info;
long	total_block_count;


static long	CrashPath[CRASH_PATH_PROFILE_LENGTH];
static long	CrashPathIndex;


void DumpDynamicData();
void DumpDynamicDataAndExit();

/*
 * Initialization
 */

void DynamicApplicationInit()
{
  long	i, j;
  char	*tmpstring;

  for (i = 0; i < CRASH_PATH_PROFILE_LENGTH; i++)
	CrashPath[i] = -1;
  CrashPathIndex = 0;

	// And install some signal catchers
  signal (SIGABRT, DumpDynamicDataAndExit);
  signal (SIGFPE, DumpDynamicDataAndExit);
  signal (SIGBUS, DumpDynamicDataAndExit);
  signal (SIGSEGV, DumpDynamicDataAndExit);
}

void DumpDynamicDataAndExit()
{
  DumpDynamicData();
  exit();
}

/*
 * Print out results
 */
void DumpDynamicData()
{
  long	i, j, path_count;

	/*
	 * Procedure counts
	 */

  fprintf (Inst_Logfile, "*** Crash Path Profile begin ***\n");
  fprintf (Inst_Logfile, "Total # Blocks executed prior to exit: %d\n", total_block_count);
  fprintf (Inst_Logfile, "Block Address\tProcedure Name\n");

	// At this point, CrashPathIndex is the next open slot...
  if (total_block_count < CRASH_PATH_PROFILE_LENGTH)
    {
    for (i = 0; i < total_block_count; i++)
	{
	long	BlockId, ProcId;

	BlockId = CrashPath[i];
	ProcId = dyn_block_info[BlockId].ProcId;
	fprintf (Inst_Logfile, "0x%lx\t%s\n",
				dyn_block_info[BlockId].Block_Addr,
				dyn_proc_count[ProcId].ProcName);
	}
    }
  else
    {
    long	BlockId, ProcId;

    i = CrashPathIndex;
    while (i != (CrashPathIndex + CRASH_PATH_PROFILE_LENGTH - 1)
			% CRASH_PATH_PROFILE_LENGTH)
	{
	BlockId = CrashPath[i];
	ProcId = dyn_block_info[BlockId].ProcId;
	fprintf (Inst_Logfile, "0x%lx\t%s\n",
				dyn_block_info[BlockId].Block_Addr,
				dyn_proc_count[ProcId].ProcName);
	i = (i + 1) % CRASH_PATH_PROFILE_LENGTH;
	}

	// And print out the last one also
    BlockId = CrashPath[i];
    ProcId = dyn_block_info[BlockId].ProcId;
    fprintf (Inst_Logfile, "0x%lx\t%s\n",
				dyn_block_info[BlockId].Block_Addr,
				dyn_proc_count[ProcId].ProcName);
    }
  fprintf(Inst_Logfile, "\n*** Crash Path Profile end ***\n");
}


/*
 * Procedure data collection
 */

void DynAllocateStorage (long Proc_Elts, long Block_Elts)
{
  Total_Procs = Proc_Elts;
  total_proc_count = 0;
  dyn_proc_count = (PROC_TABLE_TYPE *) calloc (sizeof(PROC_TABLE_TYPE), Proc_Elts);

  Total_Blocks = Block_Elts;
  total_block_count = 0;
  dyn_block_info = (BLOCK_TABLE_TYPE *) calloc (sizeof(BLOCK_TABLE_TYPE), Block_Elts);

}

void Dyn_ProcCountInit (long PSeq, char *Pname, long Addr)
{
  strncat (dyn_proc_count[PSeq].ProcName, Pname, PROC_NAME_SIZE);
  dyn_proc_count[PSeq].StartAddr = Addr;
}

void BbInfo (long this_BlockNum, long this_ProcNum,
		long this_BlockSize, long this_BlockAddr)
{
	/*
	 * Called once before the program runs, used solely for
	 * cross-correlation of stuff at data output time.
	 */
  dyn_block_info[this_BlockNum].ProcId = this_ProcNum;
  dyn_block_info[this_BlockNum].Num_Insns = this_BlockSize;
  dyn_block_info[this_BlockNum].Block_Addr = this_BlockAddr;
}

void Extend_Path (long this_BlockNum)
{
	// Amazingly simple, huh!
	// But the concept of tracking the path is the key
	// Record the current block, and simply roll forward...

  total_block_count++;
  CrashPath[CrashPathIndex] = this_BlockNum;
  CrashPathIndex = (CrashPathIndex + 1) % CRASH_PATH_PROFILE_LENGTH;
}
